#include "TMesh.h"
#include <fstream>


void TMesh::LoadBin(char *fname) {

  ifstream ifs(fname, ios::binary);
  if (ifs.fail()) {
    cerr << "INFO: cannot open file: " << fname << endl;
    return;
  }

  ifs.read((char*)&vertsN, sizeof(int));
  char yn;
  ifs.read(&yn, 1); // always xyz
  if (yn != 'y') {
    cerr << "INTERNAL ERROR: there should always be vertex xyz data" << endl;
    return;
  }
  if (verts)
    delete verts;
  verts = new V3[vertsN];

  ifs.read(&yn, 1); // cols 3 floats
  if (cols)
    delete cols;
  cols = 0;
  if (yn == 'y') {
    cols = new V3[vertsN];
  }

  ifs.read(&yn, 1); // normals 3 floats
  if (normals)
    delete normals;
  normals = 0;
  if (yn == 'y') {
    normals = new V3[vertsN];
  }

  ifs.read(&yn, 1); // texture coordinates 2 floats
  float *tcs = 0; // don't have texture coordinates for now
  if (tcs)
    delete tcs;
  tcs = 0;
  if (yn == 'y') {
    tcs = new float[vertsN*2];
  }

  ifs.read((char*)verts, vertsN*3*sizeof(float)); // load verts

  if (cols) {
    ifs.read((char*)cols, vertsN*3*sizeof(float)); // load cols
  }

  if (normals)
    ifs.read((char*)normals, vertsN*3*sizeof(float)); // load normals

  if (tcs)
    ifs.read((char*)tcs, vertsN*2*sizeof(float)); // load texture coordinates

  ifs.read((char*)&trisN, sizeof(int));
  if (tris)
    delete tris;
  tris = new unsigned int[trisN*3];
  ifs.read((char*)tris, trisN*3*sizeof(unsigned int)); // read tiangles

  ifs.close();

  cerr << "INFO: loaded " << vertsN << " verts, " << trisN << " tris from " << endl << "      " << fname << endl;
  cerr << "      xyz " << ((cols) ? "rgb " : "") << ((normals) ? "nxnynz " : "") << ((tcs) ? "tcstct " : "") << endl;

  delete []tcs;

  projVerts = new V3[vertsN];

}

void TMesh::RenderWireframe(PPC *ppc, FrameBuffer *fb) {

  if (disabled)
    return;

  for (int ti = 0; ti < trisN; ti++) {
    for (int ei = 0; ei < 3; ei++) {
      int i0 = tris[3*ti+ei];
      int i1 = tris[3*ti+(ei+1)%3];
      V3 &V0 = verts[i0];
      V3 &V1 = verts[i1];
      V3 &C0 = cols[i0];
      V3 &C1 = cols[i1];
      fb->Draw3DSegment(V0, V1, C0, C1, ppc);
    }
  }

}

void TMesh::Translate(V3 tv) {

  for (int vi = 0; vi < vertsN; vi++) {
    verts[vi] = verts[vi] + tv;
  }

}

V3 TMesh::GetCenter() {

  V3 ret(0.0f, 0.0f, 0.0f);
  for (int vi = 0; vi < vertsN; vi++) {
    ret = ret + verts[vi];
  }
  ret = ret / (float) vertsN;
  return ret;

}

void TMesh::RenderFilled(PPC *ppc, FrameBuffer *fb) {

  if (disabled)
    return;
  
  for (int vi = 0; vi < vertsN; vi++) {
    ppc->Project(verts[vi], projVerts[vi]);
  }

  for (int tri = 0; tri < trisN; tri++) {
    // vertex indices for current triangle
    int vinds[3] = {tris[3*tri+0], tris[3*tri+1], tris[3*tri+2]}; 
    if (
      projVerts[vinds[0]][0] == FLT_MAX ||
      projVerts[vinds[1]][0] == FLT_MAX ||
      projVerts[vinds[2]][0] == FLT_MAX)
      continue; // ignore triangle if any of the vertices is behind head (i.e. bad projection)

    // rasterization setup
    // screen space interpolation
    M33 ssim;
    ssim[0] = projVerts[vinds[0]];
    ssim[1] = projVerts[vinds[1]];
    ssim[2] = projVerts[vinds[2]];
    ssim.SetColumn(2, V3(1.0f, 1.0f, 1.0f));
    ssim = ssim.Inverted();

    // model space interpolation
    M33 cam; 
    cam.SetColumn(0, ppc->a);
    cam.SetColumn(1, ppc->b);
    cam.SetColumn(2, ppc->c);
    M33 tmat; 
    tmat.SetColumn(0, verts[vinds[0]]-ppc->C);
    tmat.SetColumn(1, verts[vinds[1]]-ppc->C);
    tmat.SetColumn(2, verts[vinds[2]]-ppc->C);
    M33 Q = tmat.Inverted()*cam;
    V3 msiDEF = Q[0] + Q[1] + Q[2];
    M33 msiRGBABC;
    for (int ci = 0; ci < 3; ci++) {
      V3 ccv(
        cols[vinds[0]][ci],
        cols[vinds[1]][ci],
        cols[vinds[2]][ci]
      );
      for (int j = 0; j < 3; j++) {
        msiRGBABC[ci][j] = Q.GetColumn(j)*ccv;
      }
    }

    // SSI of 1/w
    V3 wv(
      projVerts[vinds[0]][2],
      projVerts[vinds[1]][2],
      projVerts[vinds[2]][2]
    );
    V3 wABC = ssim*wv;

    // SSI of each color channel
    M33 rgbABC;
    for (int ci = 0; ci < 3; ci++) {
      V3 ccv(
        cols[vinds[0]][ci],
        cols[vinds[1]][ci],
        cols[vinds[2]][ci]
      );
      rgbABC[ci] = ssim*ccv;
    }

    // build axis aligned bounding box for projected vertices
    AABB aabb(projVerts[vinds[0]]);
    aabb.AddPoint(projVerts[vinds[1]]);
    aabb.AddPoint(projVerts[vinds[2]]);

    // clip aabb with frame

    // compute edge equations
    V3 eeqs[3]; // eeqs[ei]=(A, B, C) "Au + Bv + C"
    for (int ei = 0; ei < 3; ei++) {
      int ei1 = (ei+1)%3;
      int ei2 = (ei+2)%3;
      eeqs[ei][0] = projVerts[vinds[ei1]][1] - projVerts[vinds[ei]][1]; 
      eeqs[ei][1] = projVerts[vinds[ei]][0] - projVerts[vinds[ei1]][0]; 
      eeqs[ei][2] = 
        -projVerts[vinds[ei]][1]*eeqs[ei][1] 
        -projVerts[vinds[ei]][0]*eeqs[ei][0];
      V3 v2(projVerts[vinds[ei2]]); v2[2] = 1.0f;
      // eeqs[ei] * (u, v, 1) = Au + Bv + C
      if (eeqs[ei]*v2 < 0.0f)
        eeqs[ei] = eeqs[ei]*-1.0f;
    }

    // iterate over aabb and rasterize triangle
    int left = (int) (aabb.scc[0] + 0.5f);
    int right = (int) (aabb.lcc[0] - 0.5f);
    int top = (int) (aabb.scc[1] + 0.5f);
    int bottom = (int) (aabb.lcc[1] - 0.5f);
    for (int v = top; v <= bottom; v++) {
      for (int u = left; u <= right; u++) {
        V3 pixv(0.5f+(float)u, 0.5f+(float)v, 1.0f);

        // edge equations
        if (
          eeqs[0]*pixv < 0.0f ||
          eeqs[1]*pixv < 0.0f ||
          eeqs[2]*pixv < 0.0f)
          continue;

        // zbuffering
        float currz = wABC*pixv;
        if (fb->IsFarther(V3(pixv[0], pixv[1], currz)))
          continue;
        fb->SetZB(u, v, currz);

        // interpolate color
        // ssi
        V3 currColor;
        if (verts[0][1] > 0.0f) {
          currColor = rgbABC * pixv;
        }
        else {
          float msiDen = msiDEF * pixv;
          currColor = (msiRGBABC * pixv) / msiDen;
        }

        fb->Set(u, v, currColor.UIntColor());

      }
    }

  }

}


void TMesh::RenderHW() {

  if (disabled)
    return;

  glEnableClientState(GL_VERTEX_ARRAY); // I'm going to use an array of vertices
  glEnableClientState(GL_COLOR_ARRAY); // ... of colors

  glVertexPointer(3, GL_FLOAT, 0, verts); // here is the pointer to vertex array
  glColorPointer(3, GL_FLOAT, 0, cols); // here is the pointer to vertex array

  glDrawElements(GL_TRIANGLES, 3*trisN, GL_UNSIGNED_INT,tris); // draw these triangles

  glDisableClientState(GL_COLOR_ARRAY);
  glDisableClientState(GL_VERTEX_ARRAY);

}

void TMesh::Allocate(int _vertsN, int _trisN) {

  vertsN = _vertsN;
  trisN = _trisN;

  verts = new V3[vertsN];
  cols = new V3[vertsN];
  normals = new V3[vertsN];
  projVerts = new V3[vertsN];

  tris = new unsigned int[3*trisN];

}

void TMesh::SetQuad(V3 *_verts, V3 *_cols) {

  vertsN = 4;
  trisN = 2;
  Allocate(vertsN, trisN);
  for (int vi = 0; vi < vertsN; vi++) {
    verts[vi] = _verts[vi];
    cols[vi] = _cols[vi];
  }
  tris[0] = 0;
  tris[1] = 1;
  tris[2] = 2;

  tris[3] = 2;
  tris[4] = 3;
  tris[5] = 0;

}
